package com.bizmatics.common.spring.config.feign;

import com.bizmatics.common.core.bean.ApiResult;
import com.bizmatics.common.core.exception.ApplicationException;
import com.bizmatics.common.core.exception.BusinessException;
import com.bizmatics.common.core.exception.ErrorCode;
import com.bizmatics.common.core.util.StringUtils;
import com.bizmatics.common.spring.util.JsonUtils;
import feign.FeignException;
import feign.Response;
import feign.Util;
import feign.codec.Decoder;
import feign.codec.ErrorDecoder;
import lombok.SneakyThrows;

import java.io.IOException;
import java.lang.reflect.Type;
import java.util.Objects;

import static com.bizmatics.common.core.bean.ApiResult.ResultStatus.ERROR;

/**
 * <p>Feign 组件</p>
 *
 * @author chenpeng
 * Create at February 21, 2019 at 16:38:07 GMT+8
 */
public class FeignModuler {

    private static final String STRING_BODY_TYPE_NAME = "String";
    private static final String CALL_RESULT_STATUS_FLAG = "status";

    private static RuntimeException processException(String bodyText) {
        RuntimeException ex = null;
        if (!hasError(bodyText)) {
            return ex;
        }
        ApiResult<Void> errorResult = JsonUtils.fromJson(bodyText, ApiResult.class);
        if (errorResult == null || StringUtils.isBlank(errorResult.getCode())) {
            return ex;
        }
        String code = errorResult.getCode();
        String message = errorResult.getMsg();
        String exception = errorResult.getException();

        if (StringUtils.isBlank(exception) || exception.equals(BusinessException.class.getName())) {
            return new BusinessException(buildErrorCode(code, message));
        } else if (exception.equals(ApplicationException.class.getName())) {
            return new ApplicationException(buildErrorCode(code, message));
        }

        //实例化真正的Exception
        Class<? extends RuntimeException> exClazz;
        try {
            exClazz = (Class<? extends RuntimeException>) Class.forName(exception);
            ex = (RuntimeException) exClazz.getDeclaredConstructor(String.class)
                    .newInstance(String.format("%s - %s", code, message));
        } catch (Exception e) {
            return new BusinessException(buildErrorCode(code, message));
        }
        return ex;
    }

    private static boolean hasError(String bodyText) {
        return bodyText.contains(CALL_RESULT_STATUS_FLAG) && bodyText.contains(ERROR.name());
    }

    private static boolean isStringReturnType(Type type) {
        String typeName = type.getTypeName();
        return (STRING_BODY_TYPE_NAME.equals(typeName)
                || String.class.getName().equals(typeName));
    }

    private static ErrorCode buildErrorCode(String code, String message) {
        return new ErrorCode() {
            @Override
            public String getCode() {
                return code;
            }

            @Override
            public String getDefaultMessage() {
                return message;
            }
        };
    }

    /**
     * 解码器
     */
    static class ResultDecoder implements Decoder {
        @Override
        public Object decode(Response response, Type type) throws IOException, FeignException {
            Response.Body body = response.body();
            if (Objects.isNull(body)) {
                return null;
            }
            String bodyText = Util.toString(body.asReader());
            if (StringUtils.isBlank(bodyText)) {
                return null;
            }

            RuntimeException e = processException(bodyText);
            if (e != null) {
                throw e;
            }

            if (isStringReturnType(type)) {
                return bodyText;
            }
            return JsonUtils.fromJson(bodyText, JsonUtils.getJsonMapper().getTypeFactory().constructType(type));
        }

    }

    /**
     * 异常处理解码器
     */
    static class ExceptionDecoder implements ErrorDecoder {
        @SneakyThrows
        @Override
        public Exception decode(String s, Response response) {
            Response.Body body = response.body();
            if (Objects.isNull(body)) {
                return null;
            }
            String bodyText = Util.toString(body.asReader());
            if (StringUtils.isBlank(bodyText)) {
                return null;
            }
            return processException(bodyText);
        }
    }
}
